SystemOrganization addCategory: #'MethodMassage-Kernel'!
SystemOrganization addCategory: #'MethodMassage-Coverage'!
SystemOrganization addCategory: #'MethodMassage-Tests'!

Object subclass: #AssemblerEncoderInterface
	instanceVariableNames: 'assembler encoder'
	classVariableNames: 'SpecialSelectors'
	poolDictionaries: ''
	category: 'MethodMassage-Kernel'!

!AssemblerEncoderInterface commentStamp: 'eem 2/22/2013 13:10' prior: 0!
An AssemblerEncoderInterface is an abstract superclass for the interfaces between a BytecodeAssembler and a BytecodeEncoder.  These interfaces convert the assembly messages (see CompiledMethod>>abstractBytecodeMessages and InstructionStream interpretNextInstructionFor: et al) to the sizing and emitting messages implemented by BytecodeEncoder's subclasses.

Instance Variables
	assembler:		<BytecodeAssembler>
	encoder:		<BytecodeEncoder>!

!AssemblerEncoderInterface class methodsFor: 'instance creation' stamp: 'eem 2/22/2013 11:00'!
assembler: anAssembler encoder: anEncoder
	^self new assembler: anAssembler encoder: anEncoder! !

!AssemblerEncoderInterface class methodsFor: 'class initialization' stamp: 'eem 2/22/2013 13:29'!
copySizeCodeToEmitterCode
	"AssemblerEncoderInterface copySizeCodeToEmitterCode"
	BytecodeSizer selectorsAndMethodsDo:
		[:s :m| | emitterMethod |
		((m literals includes: #shouldBeImplemented) not
		and: [(emitterMethod := BytecodeEmitter compiledMethodAt: s) literals includes: #shouldBeImplemented]) ifTrue:
			[BytecodeEmitter
				compile:
					(m getSourceFromFile asString
						copyReplaceAll: 'encoder size'
						with: 'encoder gen')
					classified: #assembly]]! !

!AssemblerEncoderInterface class methodsFor: 'class initialization' stamp: 'eem 2/22/2013 11:31'!
createCodeSkeletons
	"BytecodeSizer createCodeSkeletons"
	"BytecodeEmitter createCodeSkeletons"
	InstructionClient selectorsAndMethodsDo:
		[:s :m|
		Parser new
			initPattern: m getSourceFromFile asString
			notifying: nil
			return:
				[:tuple|
				[:sel :args :precedence|
				self compile:
					(String streamContents:
						[:strm|
						precedence = 1
							ifTrue: [strm nextPutAll: sel]
							ifFalse:
								[sel keywords with: args do:
									[:kw :arg|
									strm nextPutAll: kw; space; nextPutAll: arg; space].
								 strm skip: -1].
						strm crtab; nextPutAll: 'self shouldBeImplemented'])
					classified: #assembly.
				] valueWithArguments: tuple]]! !

!AssemblerEncoderInterface class methodsFor: 'class initialization' stamp: 'eem 2/22/2013 13:20'!
initialize
	"AssemblerEncoderInterface initialize"
	SpecialSelectors := IdentityDictionary new.
	1 to: Smalltalk specialSelectorSize do:
		[:i | SpecialSelectors at: (Smalltalk specialSelectorAt: i) put: i]! !

!AssemblerEncoderInterface methodsFor: 'accessing' stamp: 'eem 2/22/2013 10:59'!
assembler
	"Answer the value of assembler"

	^ assembler! !

!AssemblerEncoderInterface methodsFor: 'accessing' stamp: 'eem 2/22/2013 10:59'!
assembler: anObject
	"Set the value of assembler"

	assembler := anObject! !

!AssemblerEncoderInterface methodsFor: 'initialize-release' stamp: 'eem 2/22/2013 11:01'!
assembler: anAssembler encoder: anEncoder
	assembler := anAssembler.
	encoder := anEncoder! !

!AssemblerEncoderInterface methodsFor: 'accessing' stamp: 'eem 2/22/2013 10:59'!
encoder
	"Answer the value of encoder"

	^ encoder! !

!AssemblerEncoderInterface methodsFor: 'accessing' stamp: 'eem 2/22/2013 10:59'!
encoder: anObject
	"Set the value of encoder"

	encoder := anObject! !

AssemblerEncoderInterface subclass: #BytecodeEmitter
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MethodMassage-Kernel'!

!BytecodeEmitter commentStamp: 'eem 2/22/2013 11:18' prior: 0!
A BytecodeEmitter converts the assembly messages (see CompiledMethod>>abstractBytecodeMessages and InstructionStream interpretNextInstructionFor: et al) and sends them to the corresponding emitting messages implemented by BytecodeEncoder's subclasses.

Instance Variables: see superclass!

!BytecodeEmitter methodsFor: 'assembly' stamp: 'eem 8/1/2014 07:13'!
blockReturnConstant: aLiteral
	"Rubbish way of handling special constants."
	^[encoder genReturnSpecialLiteralToCaller: aLiteral]
		on: Error
		do: [:ex| ex return: (self pushLiteral: aLiteral) + self blockReturnTop]! !

!BytecodeEmitter methodsFor: 'assembly' stamp: 'eem 2/22/2013 15:27'!
blockReturnTop
	^encoder genReturnTopToCaller! !

!BytecodeEmitter methodsFor: 'assembly' stamp: 'eem 2/23/2013 21:36'!
callPrimitive: primitiveIndex
	^encoder genCallPrimitive: primitiveIndex! !

!BytecodeEmitter methodsFor: 'assembly' stamp: 'eem 2/23/2013 17:47'!
doDup
	^encoder genDup! !

!BytecodeEmitter methodsFor: 'assembly' stamp: 'eem 2/22/2013 13:29'!
doPop
	^encoder genPop! !

!BytecodeEmitter methodsFor: 'assembly' stamp: 'eem 3/15/2013 14:04'!
extA: anInteger 
	^encoder genUnsignedSingleExtendA: anInteger! !

!BytecodeEmitter methodsFor: 'assembly' stamp: 'eem 3/15/2013 14:04'!
extB: anInteger 
	^encoder genUnsignedSingleExtendB: anInteger! !

!BytecodeEmitter methodsFor: 'assembly' stamp: 'eem 2/22/2013 14:34'!
jump: offsetOrLabel
	| offset |
	offset := (assembler isLabel: offsetOrLabel)
				ifTrue: [assembler spanToLabel: offsetOrLabel]
				ifFalse: [offsetOrLabel].
	^encoder genJump: offset! !

!BytecodeEmitter methodsFor: 'assembly' stamp: 'eem 2/22/2013 15:10'!
jump: offsetOrLabel if: condition
	| offset |
	offset := (assembler isLabel: offsetOrLabel)
				ifTrue: [assembler spanToLabel: offsetOrLabel]
				ifFalse: [offsetOrLabel].
	condition
		ifTrue: [encoder genBranchPopTrue: offset]
		ifFalse: [encoder genBranchPopFalse: offset]! !

!BytecodeEmitter methodsFor: 'assembly' stamp: 'eem 3/15/2013 15:47'!
jumpLong: offsetOrLabel
	| offset |
	offset := (assembler isLabel: offsetOrLabel)
				ifTrue: [assembler spanToLabel: offsetOrLabel]
				ifFalse: [offsetOrLabel].
	^encoder genJumpLong: offset! !

!BytecodeEmitter methodsFor: 'assembly' stamp: 'eem 3/15/2013 15:47'!
jumpLong: offsetOrLabel if: condition
	| offset |
	offset := (assembler isLabel: offsetOrLabel)
				ifTrue: [assembler spanToLabel: offsetOrLabel]
				ifFalse: [offsetOrLabel].
	^condition
		ifTrue: [encoder genBranchPopTrueLong: offset]
		ifFalse: [encoder genBranchPopFalseLong: offset]! !

!BytecodeEmitter methodsFor: 'assembly' stamp: 'eem 2/23/2013 16:28'!
methodReturnConstant: aLiteral
	"Rubbish way of handling special constants."
	^[encoder genReturnSpecialLiteral: aLiteral]
		on: Error
		do: [:ex| ex return: (self pushLiteral: aLiteral) + self methodReturnTop]! !

!BytecodeEmitter methodsFor: 'assembly' stamp: 'eem 2/22/2013 15:27'!
methodReturnReceiver
	^encoder genReturnReceiver! !

!BytecodeEmitter methodsFor: 'assembly' stamp: 'eem 2/22/2013 13:29'!
methodReturnTop
	^encoder genReturnTop! !

!BytecodeEmitter methodsFor: 'assembly' stamp: 'eem 2/23/2013 18:03'!
popIntoLiteralVariable: aLiteral
	| litIndex |
	litIndex := encoder litIndexOrNilFor: aLiteral.
	litIndex ifNil:
		[litIndex := encoder litIndex: aLiteral].
	^encoder genStorePopLiteralVar: litIndex! !

!BytecodeEmitter methodsFor: 'assembly' stamp: 'eem 3/14/2013 16:46'!
popIntoLiteralVariableAtIndex: litIndex
	^encoder genStorePopLiteralVar: litIndex! !

!BytecodeEmitter methodsFor: 'assembly' stamp: 'eem 2/24/2013 11:42'!
popIntoReceiverVariable: offsetOrName
	| offset |
	offset := offsetOrName isInteger
				ifTrue: [offsetOrName]
				ifFalse: [assembler offsetForInstVarName: offsetOrName].
	"handle context inst var access"
	^(assembler shouldUseLongAccessForInstVarOffset: offset)
		ifTrue: [encoder genStorePopInstVarLong: offset]
		ifFalse: [encoder genStorePopInstVar: offset]! !

!BytecodeEmitter methodsFor: 'assembly' stamp: 'eem 2/23/2013 17:54'!
popIntoRemoteTemp: remoteTempIndex inVectorAt: tempVectorIndex
	^encoder genStorePopRemoteTemp: remoteTempIndex inVectorAt: tempVectorIndex! !

!BytecodeEmitter methodsFor: 'assembly' stamp: 'eem 2/22/2013 15:09'!
popIntoTemporaryVariable: offset
	encoder genStorePopTemp: offset! !

!BytecodeEmitter methodsFor: 'assembly' stamp: 'eem 2/23/2013 17:47'!
pushActiveContext
	^encoder genPushThisContext! !

!BytecodeEmitter methodsFor: 'assembly' stamp: 'eem 2/22/2013 15:25'!
pushClosureCopyNumCopiedValues: numCopied numArgs: numArgs blockSize: blockSizeOrLabel
	| blockSize |
	blockSize := (assembler isLabel: blockSizeOrLabel)
					ifTrue: [assembler spanToLabel: blockSizeOrLabel]
					ifFalse: [blockSizeOrLabel].
	^encoder genPushClosureCopyNumCopiedValues: numCopied numArgs: numArgs jumpSize: blockSize! !

!BytecodeEmitter methodsFor: 'assembly' stamp: 'eem 2/23/2013 17:51'!
pushConsArrayWithElements: numElements
	^encoder genPushConsArray: numElements! !

!BytecodeEmitter methodsFor: 'assembly' stamp: 'eem 2/22/2013 16:36'!
pushConstant: aLiteral
	| litIndex |
	litIndex := encoder litIndexOrNilFor: aLiteral.
	litIndex ifNil:
		["rubbish way of handling special literals"
		 ([encoder genPushSpecialLiteral: aLiteral]
				on: Error
				do: [:ex| ex return: nil]) ifNotNil:
			[^self].
		 litIndex := encoder litIndex: aLiteral].
	^encoder genPushLiteral: litIndex! !

!BytecodeEmitter methodsFor: 'assembly' stamp: 'eem 3/14/2013 16:40'!
pushConstantAtIndex: literalIndex
	^encoder genPushLiteral: literalIndex! !

!BytecodeEmitter methodsFor: 'assembly' stamp: 'eem 2/22/2013 14:10'!
pushLiteralVariable: aLiteral
	| litIndex |
	litIndex := encoder litIndexOrNilFor: aLiteral.
	litIndex ifNil:
		[litIndex := encoder litIndex: aLiteral].
	^encoder genPushLiteralVar: litIndex! !

!BytecodeEmitter methodsFor: 'assembly' stamp: 'eem 3/14/2013 16:47'!
pushLiteralVariableAtIndex: litIndex
	^encoder genPushLiteralVar: litIndex! !

!BytecodeEmitter methodsFor: 'assembly' stamp: 'eem 2/23/2013 17:50'!
pushNewArrayOfSize: numElements
	^encoder genPushNewArray: numElements! !

!BytecodeEmitter methodsFor: 'assembly' stamp: 'eem 2/22/2013 13:29'!
pushReceiver
	^encoder genPushReceiver! !

!BytecodeEmitter methodsFor: 'assembly' stamp: 'eem 2/24/2013 11:42'!
pushReceiverVariable: offsetOrName
	| offset |
	offset := offsetOrName isInteger
				ifTrue: [offsetOrName]
				ifFalse: [assembler offsetForInstVarName: offsetOrName].
	"handle context inst var access"
	^(assembler shouldUseLongAccessForInstVarOffset: offset)
		ifTrue: [encoder genPushInstVarLong: offset]
		ifFalse: [encoder genPushInstVar: offset]! !

!BytecodeEmitter methodsFor: 'assembly' stamp: 'eem 2/23/2013 17:49'!
pushRemoteTemp: remoteTempIndex inVectorAt: tempVectorIndex
	^encoder genPushRemoteTemp: remoteTempIndex inVectorAt: tempVectorIndex! !

!BytecodeEmitter methodsFor: 'assembly' stamp: 'eem 2/23/2013 16:49'!
pushSpecialConstant: aLiteral 
	^encoder genPushSpecialLiteral: aLiteral! !

!BytecodeEmitter methodsFor: 'assembly' stamp: 'eem 2/22/2013 13:29'!
pushTemporaryVariable: offset
	^encoder genPushTemp: offset! !

!BytecodeEmitter methodsFor: 'assembly' stamp: 'eem 2/22/2013 13:31'!
send: selector super: supered numArgs: nArgs
	supered ifFalse:
		[(SpecialSelectors at: selector ifAbsent: []) ifNotNil:
			[:specialSelectorIndex|
			 ^encoder genSendSpecial: specialSelectorIndex numArgs: nArgs].
		^encoder genSend: (encoder sharableLitIndex: selector) numArgs: nArgs].
	^encoder genSendSuper: (encoder sharableLitIndex: selector) numArgs: nArgs! !

!BytecodeEmitter methodsFor: 'assembly' stamp: 'eem 3/22/2013 15:25'!
sendToAbsentDynamicSuperclass: selector numArgs: numArgs
	^encoder genSendAbsentDynamicSuper: (encoder sharableLitIndex: selector) numArgs: numArgs! !

!BytecodeEmitter methodsFor: 'assembly' stamp: 'eem 3/15/2013 13:46'!
sendToAbsentImplicitReceiver: selector numArgs: numArgs
	^encoder genSendAbsentImplicit: (encoder sharableLitIndex: selector) numArgs: numArgs! !

!BytecodeEmitter methodsFor: 'assembly' stamp: 'eem 2/23/2013 21:40'!
storeIntoLiteralVariable: aLiteral
	| litIndex |
	litIndex := encoder litIndexOrNilFor: aLiteral.
	litIndex ifNil:
		[litIndex := encoder litIndex: aLiteral].
	^encoder genStoreLiteralVar: litIndex! !

!BytecodeEmitter methodsFor: 'assembly' stamp: 'eem 3/14/2013 16:46'!
storeIntoLiteralVariableAtIndex: litIndex
	^encoder genStoreLiteralVar: litIndex! !

!BytecodeEmitter methodsFor: 'assembly' stamp: 'eem 2/24/2013 11:41'!
storeIntoReceiverVariable: offsetOrName
	| offset |
	offset := offsetOrName isInteger
				ifTrue: [offsetOrName]
				ifFalse: [assembler offsetForInstVarName: offsetOrName].
	"handle context inst var access"
	^(assembler shouldUseLongAccessForInstVarOffset: offset)
		ifTrue: [encoder genStoreInstVarLong: offset]
		ifFalse: [encoder genStoreInstVar: offset]! !

!BytecodeEmitter methodsFor: 'assembly' stamp: 'eem 2/23/2013 17:49'!
storeIntoRemoteTemp: remoteTempIndex inVectorAt: tempVectorIndex
	^encoder genStoreRemoteTemp: remoteTempIndex inVectorAt: tempVectorIndex! !

!BytecodeEmitter methodsFor: 'assembly' stamp: 'eem 2/22/2013 16:43'!
storeIntoTemporaryVariable: offset
	^encoder genStoreTemp: offset! !

!BytecodeEmitter methodsFor: 'assembly' stamp: 'eem 8/1/2014 20:43'!
trapIfNotInstanceOf: behaviorOrArrayOfBehavior
	| litIndex |
	litIndex := encoder litIndexOrNilFor: behaviorOrArrayOfBehavior.
	litIndex ifNil:
		[litIndex := encoder litIndex: behaviorOrArrayOfBehavior].
	^encoder genTrapIfNotInstanceOf: litIndex! !

AssemblerEncoderInterface subclass: #BytecodeSizer
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MethodMassage-Kernel'!

!BytecodeSizer commentStamp: 'eem 2/22/2013 11:18' prior: 0!
A BytecodeSizer converts the assembly messages (see CompiledMethod>>abstractBytecodeMessages and InstructionStream interpretNextInstructionFor: et al) and sends them to the corresponding sizing messages implemented by BytecodeEncoder's subclasses.

Instance Variables: see superclass!

!BytecodeSizer methodsFor: 'assembly' stamp: 'eem 8/1/2014 07:14'!
blockReturnConstant: aLiteral
	"Rubbish way of handling special constants."
	^[encoder sizeReturnSpecialLiteralToCaller: aLiteral]
		on: Error
		do: [:ex| ex return: (self pushLiteral: aLiteral) + self blockReturnTop]! !

!BytecodeSizer methodsFor: 'assembly' stamp: 'eem 2/22/2013 15:26'!
blockReturnTop
	^encoder sizeReturnTopToCaller! !

!BytecodeSizer methodsFor: 'assembly' stamp: 'eem 2/23/2013 21:35'!
callPrimitive: primitiveIndex
	^encoder sizeCallPrimitive: primitiveIndex! !

!BytecodeSizer methodsFor: 'assembly' stamp: 'eem 2/23/2013 17:46'!
doDup
	^encoder sizeDup! !

!BytecodeSizer methodsFor: 'assembly' stamp: 'eem 2/22/2013 13:21'!
doPop
	^encoder sizePop! !

!BytecodeSizer methodsFor: 'assembly' stamp: 'eem 3/15/2013 14:01'!
extA: anInteger 
	^encoder sizeUnsignedSingleExtendA: anInteger! !

!BytecodeSizer methodsFor: 'assembly' stamp: 'eem 3/15/2013 14:00'!
extB: anInteger 
	^encoder sizeUnsignedSingleExtendB: anInteger! !

!BytecodeSizer methodsFor: 'assembly' stamp: 'eem 2/22/2013 15:02'!
jump: offsetOrLabel
	| offset |
	offset := (assembler isLabel: offsetOrLabel)
				ifTrue: [assembler spanToLabel: offsetOrLabel]
				ifFalse: [offsetOrLabel].
	^offset
		ifNotNil: [encoder sizeJump: offset]
		ifNil: ["If this is a backward jump that is as-yet unsized, answer the long backward jump size.
			   c.f. e.g. MessageNode>>#sizeCodeForWhile:value:"
			  offsetOrLabel value <= assembler index ifTrue:
				[encoder sizeJumpLong: -1]]! !

!BytecodeSizer methodsFor: 'assembly' stamp: 'eem 2/22/2013 14:21'!
jump: offsetOrLabel if: condition
	| offset |
	offset := (assembler isLabel: offsetOrLabel)
				ifTrue: [assembler spanToLabel: offsetOrLabel]
				ifFalse: [offsetOrLabel].
	^offset ifNotNil:
		[condition
			ifTrue: [encoder sizeBranchPopTrue: offset]
			ifFalse: [encoder sizeBranchPopFalse: offset]]! !

!BytecodeSizer methodsFor: 'assembly' stamp: 'eem 3/15/2013 15:48'!
jumpLong: offsetOrLabel
	| offset |
	offset := (assembler isLabel: offsetOrLabel)
				ifTrue: [assembler spanToLabel: offsetOrLabel]
				ifFalse: [offsetOrLabel].
	^offset ifNotNil:
		[encoder sizeJumpLong: offset]! !

!BytecodeSizer methodsFor: 'assembly' stamp: 'eem 3/15/2013 15:38'!
jumpLong: offsetOrLabel if: condition
	| offset |
	offset := (assembler isLabel: offsetOrLabel)
				ifTrue: [assembler spanToLabel: offsetOrLabel]
				ifFalse: [offsetOrLabel].
	^offset ifNotNil:
		[condition
			ifTrue: [encoder sizeBranchPopTrueLong: offset]
			ifFalse: [encoder sizeBranchPopFalseLong: offset]]! !

!BytecodeSizer methodsFor: 'assembly' stamp: 'eem 2/23/2013 16:27'!
methodReturnConstant: aLiteral
	"Rubbish way of handling special constants."
	^[encoder sizeReturnSpecialLiteral: aLiteral]
		on: Error
		do: [:ex| ex return: (self pushLiteral: aLiteral) + self methodReturnTop]! !

!BytecodeSizer methodsFor: 'assembly' stamp: 'eem 2/22/2013 15:26'!
methodReturnReceiver
	^encoder sizeReturnReceiver! !

!BytecodeSizer methodsFor: 'assembly' stamp: 'eem 2/22/2013 13:21'!
methodReturnTop
	^encoder sizeReturnTop! !

!BytecodeSizer methodsFor: 'assembly' stamp: 'eem 2/23/2013 18:03'!
popIntoLiteralVariable: aLiteral
	| litIndex |
	litIndex := encoder litIndexOrNilFor: aLiteral.
	litIndex ifNil:
		[litIndex := encoder litIndex: aLiteral].
	^encoder sizeStorePopLiteralVar: litIndex! !

!BytecodeSizer methodsFor: 'assembly' stamp: 'eem 3/14/2013 16:47'!
popIntoLiteralVariableAtIndex: litIndex
	^encoder sizeStorePopLiteralVar: litIndex! !

!BytecodeSizer methodsFor: 'assembly' stamp: 'eem 2/24/2013 11:40'!
popIntoReceiverVariable: offsetOrName
	| offset |
	offset := offsetOrName isInteger
				ifTrue: [offsetOrName]
				ifFalse: [assembler offsetForInstVarName: offsetOrName].
	"handle context inst var access"
	^(assembler shouldUseLongAccessForInstVarOffset: offset)
		ifTrue: [encoder sizeStorePopInstVarLong: offset]
		ifFalse: [encoder sizeStorePopInstVar: offset]! !

!BytecodeSizer methodsFor: 'assembly' stamp: 'eem 2/23/2013 17:51'!
popIntoRemoteTemp: remoteTempIndex inVectorAt: tempVectorIndex
	^encoder sizePushRemoteTemp: remoteTempIndex inVectorAt: tempVectorIndex! !

!BytecodeSizer methodsFor: 'assembly' stamp: 'eem 2/22/2013 14:14'!
popIntoTemporaryVariable: offset
	^encoder sizeStorePopTemp: offset! !

!BytecodeSizer methodsFor: 'assembly' stamp: 'eem 2/23/2013 17:47'!
pushActiveContext
	^encoder sizePushThisContext! !

!BytecodeSizer methodsFor: 'assembly' stamp: 'eem 2/22/2013 15:24'!
pushClosureCopyNumCopiedValues: numCopied numArgs: numArgs blockSize: blockSizeOrLabel
	| blockSize |
	blockSize := (assembler isLabel: blockSizeOrLabel)
					ifTrue: [assembler spanToLabel: blockSizeOrLabel]
					ifFalse: [blockSizeOrLabel].
	^blockSize ifNotNil:
		[encoder sizePushClosureCopyNumCopiedValues: numCopied numArgs: numArgs jumpSize: blockSize]! !

!BytecodeSizer methodsFor: 'assembly' stamp: 'eem 2/23/2013 17:51'!
pushConsArrayWithElements: numElements
	^encoder sizePushConsArray: numElements! !

!BytecodeSizer methodsFor: 'assembly' stamp: 'eem 2/22/2013 16:28'!
pushConstant: aLiteral
	| litIndex |
	litIndex := encoder litIndexOrNilFor: aLiteral.
	litIndex ifNil:
		["rubbish way of handling special literals"
		 [(encoder sizePushSpecialLiteral: aLiteral) ifNotNil:
			[:size| ^size]]
			on: Error
			do: [:ex| ].
		 litIndex := encoder litIndex: aLiteral].
	^encoder sizePushLiteral: litIndex! !

!BytecodeSizer methodsFor: 'assembly' stamp: 'eem 3/14/2013 16:44'!
pushConstantAtIndex: literalIndex
	^encoder sizePushLiteral: literalIndex! !

!BytecodeSizer methodsFor: 'assembly' stamp: 'eem 2/22/2013 14:09'!
pushLiteralVariable: aLiteral
	| litIndex |
	litIndex := encoder litIndexOrNilFor: aLiteral.
	litIndex ifNil:
		[litIndex := encoder litIndex: aLiteral].
	^encoder sizePushLiteralVar: litIndex! !

!BytecodeSizer methodsFor: 'assembly' stamp: 'eem 3/14/2013 16:47'!
pushLiteralVariableAtIndex: litIndex
	^encoder sizePushLiteralVar: litIndex! !

!BytecodeSizer methodsFor: 'assembly' stamp: 'eem 2/23/2013 17:48'!
pushNewArrayOfSize: numElements
	^encoder sizePushNewArray: numElements! !

!BytecodeSizer methodsFor: 'assembly' stamp: 'eem 2/22/2013 13:20'!
pushReceiver
	^encoder sizePushReceiver! !

!BytecodeSizer methodsFor: 'assembly' stamp: 'eem 2/24/2013 11:40'!
pushReceiverVariable: offsetOrName
	| offset |
	offset := offsetOrName isInteger
				ifTrue: [offsetOrName]
				ifFalse: [assembler offsetForInstVarName: offsetOrName].
	"handle context inst var access"
	^(assembler shouldUseLongAccessForInstVarOffset: offset)
		ifTrue: [encoder sizePushInstVarLong: offset]
		ifFalse: [encoder sizePushInstVar: offset]! !

!BytecodeSizer methodsFor: 'assembly' stamp: 'eem 2/23/2013 17:49'!
pushRemoteTemp: remoteTempIndex inVectorAt: tempVectorIndex
	^encoder sizePushRemoteTemp: remoteTempIndex inVectorAt: tempVectorIndex! !

!BytecodeSizer methodsFor: 'assembly' stamp: 'eem 2/23/2013 16:49'!
pushSpecialConstant: aLiteral 
	^encoder sizePushSpecialLiteral: aLiteral! !

!BytecodeSizer methodsFor: 'assembly' stamp: 'eem 2/22/2013 12:50'!
pushTemporaryVariable: offset
	^encoder sizePushTemp: offset! !

!BytecodeSizer methodsFor: 'assembly' stamp: 'eem 2/22/2013 13:16'!
send: selector super: supered numArgs: nArgs
	supered ifFalse:
		[(SpecialSelectors at: selector ifAbsent: []) ifNotNil:
			[:specialSelectorIndex|
			 ^encoder sizeSpecialSend: specialSelectorIndex numArgs: nArgs].
		^encoder sizeSend: (encoder sharableLitIndex: selector) numArgs: nArgs].
	^encoder sizeSendSuper: (encoder sharableLitIndex: selector) numArgs: nArgs! !

!BytecodeSizer methodsFor: 'assembly' stamp: 'eem 3/22/2013 15:30'!
sendToAbsentDynamicSuperclass: selector numArgs: numArgs
	^encoder sizeSendAbsentDynamicSuper: (encoder sharableLitIndex: selector) numArgs: numArgs! !

!BytecodeSizer methodsFor: 'assembly' stamp: 'eem 3/15/2013 13:42'!
sendToAbsentImplicitReceiver: selector numArgs: numArgs
	^encoder sizeSendAbsentImplicit: (encoder sharableLitIndex: selector) numArgs: numArgs! !

!BytecodeSizer methodsFor: 'assembly' stamp: 'eem 2/23/2013 21:40'!
storeIntoLiteralVariable: aLiteral
	| litIndex |
	litIndex := encoder litIndexOrNilFor: aLiteral.
	litIndex ifNil:
		[litIndex := encoder litIndex: aLiteral].
	^encoder sizeStoreLiteralVar: litIndex! !

!BytecodeSizer methodsFor: 'assembly' stamp: 'eem 3/14/2013 16:45'!
storeIntoLiteralVariableAtIndex: litIndex
	^encoder sizeStoreLiteralVar: litIndex! !

!BytecodeSizer methodsFor: 'assembly' stamp: 'eem 2/24/2013 11:40'!
storeIntoReceiverVariable: offsetOrName
	| offset |
	offset := offsetOrName isInteger
				ifTrue: [offsetOrName]
				ifFalse: [assembler offsetForInstVarName: offsetOrName].
	"handle context inst var access"
	^(assembler shouldUseLongAccessForInstVarOffset: offset)
		ifTrue: [encoder sizeStoreInstVarLong: offset]
		ifFalse: [encoder sizeStoreInstVar: offset]! !

!BytecodeSizer methodsFor: 'assembly' stamp: 'eem 2/23/2013 17:48'!
storeIntoRemoteTemp: remoteTempIndex inVectorAt: tempVectorIndex
	^encoder sizeStoreRemoteTemp: remoteTempIndex inVectorAt: tempVectorIndex! !

!BytecodeSizer methodsFor: 'assembly' stamp: 'eem 2/22/2013 16:42'!
storeIntoTemporaryVariable: offset
	^encoder sizeStoreTemp: offset! !

!BytecodeSizer methodsFor: 'assembly' stamp: 'eem 8/1/2014 20:42'!
trapIfNotInstanceOf: behaviorOrArrayOfBehavior
	| litIndex |
	litIndex := encoder litIndexOrNilFor: behaviorOrArrayOfBehavior.
	litIndex ifNil:
		[litIndex := encoder litIndex: behaviorOrArrayOfBehavior].
	^encoder sizeTrapIfNotInstanceOf: litIndex! !

Object subclass: #AssemblerMethod
	instanceVariableNames: 'literals instructions numArgs numTemps frameSize primitive flag signFlag trailer methodClass selector compiledMethodClass fixLabels encoder'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MethodMassage-Kernel'!

!AssemblerMethod commentStamp: 'eem 8/7/2014 11:51' prior: 0!
An AssemblerMethod is a holder for a sequence of literals and a sequence of instructions which can represent a disassembled method, or a method being assembled.  AssemblerMethod supports three modes of operation.

The first mode is transcription, CompiledMethod -> AssemblerMethod -> CompiledMethod.
e.g.
	| original "CompiledMethod"
	  assembler "AssemblerMethod"
	  replica "CompiledMethod" |
	original := Object >> #printOn:.
	assembler := original disassemble.
	replica := assembler assemble.
	self assert: original = replica

The second mode is to construct an AssemblerMethod and assemble it.  e.g.
	EncoderForSistaV1 assembler
		pushReceiver;
		trapIfNotInstanceOf: {SmallInteger. Character. LargePositiveInteger. LargeNegativeInteger.
							   Integer. Fraction. Magnitude};
		pushReceiver;
		methodReturnTop;
		assemble

	AssemblerMethod new
		methodClass: Object;
		selector: #printOn:;
		numArgs: 1;
		numTemps: 2;
		pushReceiver;
		send: #class;
		send: #name;
		popIntoTemporaryVariable: 1;
		pushTemporaryVariable: 0;
		doDup;
		pushTemporaryVariable: 1;
		send: #first;
		send: #isVowel;
		jump: 'L1' if: false;
		pushConstant: 'an ';
		jump: 'L2';
		label: 'L1';
		pushConstant: 'a ';
		label: 'L2';
		send: #nextPutAll:;
		doPop;
		pushTemporaryVariable: 1;
		send: #nextPutAll:;
		doPop;
		methodReturnReceiver;
		assemble

A third mode, useful for storage, inter-image transporation ,etc, is AssemblerMethod -> executable string -> AssemblerMethod.
e.g.
	| original "CompiledMethod"
	  assembler "AssemblerMethod"
	  assemblerText "String"
	  replica "CompiledMethod" |
	original := Object >> #printOn:.
	assembler := original disassemble.
	assemblerText := assembler assemblerString.
	replica := (Compiler evaluate: assemblerText) assemble.
	self assert: original = replica

Instance Variables
	fixLabels:		<UndefinedObject|Boolean>
	flag:			<Boolean>
	frameSize:		<Integer>
	instructions:	<SequenceableCollection>
	literals:			<SequenceableCollection>
	methodClass:	<Behavior>
	numArgs:		<Integer>
	numTemps:		<Integer>
	primitive:		<Integer>
	signFlag:		<Boolean>
	trailer:			<CompiledMethodTrailer>

fixLabels
	- if nil, the AssemblerMethod is being used in transcription mode and branch offsets should be integers.
	  if a boolean then it indicates whether instructions contains one or more instructions that need their branches fixing up.

flag
	- the value of the method's header flag (see CompiledMethod>>flag)

frameSize
	- the size of the method's frame (see CompiledMethod>>frameSize)

instructions
	- the sequence of Message and LookupKey objects making up the method's instructions

literals
	- the sequence of objects making up the method's literals

methodClass
	- the target class for the method (see CompiledMethod>>methodClass)

numArgs
	- the method's argument count (see CompiledMethod>>numArgs)

numTemps
	- the method's temporary count (see CompiledMethod>>numTemps)

primitive
	- the method's primitive (see CompiledMethod>>primitive)

signFlag
	- the value of the method's signFlag (see CompiledMethod>>signFlag)

trailer
	- the method;s trailer (see CompiledMethod>>trailer)
!

!AssemblerMethod class methodsFor: 'instance creation' stamp: 'eem 2/24/2013 13:18'!
literals: literalSequence instructions: instructionSequence numArgs: nArgs numTemps: nTemps frameSize: frameSizeNum primitive: primitiveIndex flag: flagBoolean signFlag: signFlagBoolean trailer: methodTrailer methodClass: class selector: aSelector compiledMethodClass: classOfCompiledMethod
	^self new literals: literalSequence instructions: instructionSequence numArgs: nArgs numTemps: nTemps frameSize: frameSizeNum primitive: primitiveIndex flag: flagBoolean signFlag: signFlagBoolean trailer: methodTrailer methodClass: class selector: aSelector compiledMethodClass: classOfCompiledMethod! !

!AssemblerMethod methodsFor: 'assembling' stamp: 'eem 8/7/2014 11:10'!
assemble
	| assembler |
	assembler := BytecodeAssembler new.
	encoder ifNotNil: [assembler encoder: encoder].
	^assembler assemble: self! !

!AssemblerMethod methodsFor: 'printing' stamp: 'eem 3/15/2013 13:18'!
assemblerString
	^String streamContents: [:s| self printAsAssemblerOn: s]! !

!AssemblerMethod methodsFor: 'assembling' stamp: 'eem 3/14/2013 15:53'!
binding: aSymbol
	self literal: (methodClass bindingOf: aSymbol)! !

!AssemblerMethod methodsFor: 'private' stamp: 'eem 3/15/2013 16:06'!
checkLabel: label
	fixLabels ifNotNil:
		[label isInteger
			ifTrue: [self error: 'one cannot use integers as labels in assembler mode. integers are interpreted as branch distances']
			ifFalse: [fixLabels := true]]! !

!AssemblerMethod methodsFor: 'accessing' stamp: 'eem 2/22/2013 14:06'!
compiledMethodClass
	"Answer the value of compiledMethodClass"

	^ compiledMethodClass! !

!AssemblerMethod methodsFor: 'accessing' stamp: 'eem 2/22/2013 14:06'!
compiledMethodClass: anObject
	"Set the value of compiledMethodClass"

	compiledMethodClass := anObject! !

!AssemblerMethod methodsFor: 'assembling' stamp: 'eem 8/7/2014 11:25'!
doesNotUnderstand: aMessage
	(InstructionClient includesSelector: aMessage selector) ifFalse:
		[^super doesNotUnderstand: aMessage].
	aMessage lookupClass: nil.
	instructions addLast: aMessage! !

!AssemblerMethod methodsFor: 'accessing' stamp: 'eem 8/7/2014 11:07'!
encoder
	^encoder! !

!AssemblerMethod methodsFor: 'accessing' stamp: 'eem 8/7/2014 11:08'!
encoder: anObject
	encoder := anObject! !

!AssemblerMethod methodsFor: 'assembling' stamp: 'eem 08/07/2014 11:24'!
extA: anInteger 
	instructions addLast: (Message selector: #extA: argument: anInteger)! !

!AssemblerMethod methodsFor: 'assembling' stamp: 'eem 08/07/2014 11:24'!
extB: anInteger 
	instructions addLast: (Message selector: #extB: argument: anInteger)! !

!AssemblerMethod methodsFor: 'private' stamp: 'eem 3/15/2013 16:10'!
fixLabelsIfNecessary
	| map |
	fixLabels ~~ true ifTrue: [^self].
	map := Dictionary new.
	instructions do:
		[:inst|
		(self isLabel: inst) ifTrue:
			[map at: inst key put: inst]].
	instructions do:
		[:inst|
		((self isLabel: inst) not
		 and: [#(jump: jump:if: jumpLong: jumpLong:if:
				pushClosureCopyNumCopiedValues:numArgs:blockSize:) includes: inst selector]) ifTrue:
			[inst selector first = $j
				ifTrue:
					[inst arguments first isInteger ifFalse:
						[inst arguments at: 1 put: (map at: inst arguments first)]]
				ifFalse:
					[inst arguments last isInteger ifFalse:
						[inst arguments at: inst numArgs put: (map at: inst arguments last)]]]].
	fixLabels := false! !

!AssemblerMethod methodsFor: 'accessing' stamp: 'eem 2/22/2013 04:31'!
flag
	"Answer the value of flag"

	^ flag! !

!AssemblerMethod methodsFor: 'accessing' stamp: 'eem 3/14/2013 14:39'!
flag: anObject
	"Set the value of flag"

	flag := anObject! !

!AssemblerMethod methodsFor: 'accessing' stamp: 'eem 2/22/2013 04:31'!
frameSize
	"Answer the value of frameSize"

	^ frameSize! !

!AssemblerMethod methodsFor: 'accessing' stamp: 'eem 2/22/2013 04:31'!
frameSize: anObject
	"Set the value of frameSize"

	frameSize := anObject! !

!AssemblerMethod methodsFor: 'initialize-release' stamp: 'eem 8/1/2014 20:50'!
initialize
	fixLabels := false.
	compiledMethodClass := CompiledMethod.
	instructions := OrderedCollection new.
	literals := OrderedCollection new.
	frameSize := thisContext method frameSize. "should be small"
	flag := signFlag := false.
	numArgs := numTemps := primitive := 0! !

!AssemblerMethod methodsFor: 'accessing' stamp: 'eem 3/15/2013 10:43'!
instructions
	"Answer the value of instructions"
	self fixLabelsIfNecessary.
	^instructions! !

!AssemblerMethod methodsFor: 'accessing' stamp: 'eem 2/22/2013 04:31'!
instructions: anObject
	"Set the value of instructions"

	instructions := anObject! !

!AssemblerMethod methodsFor: 'private' stamp: 'eem 3/15/2013 10:26'!
isLabel: aMessageOrLookupKey
	^aMessageOrLookupKey isVariableBinding! !

!AssemblerMethod methodsFor: 'assembling' stamp: 'eem 08/07/2014 11:24'!
jump: label
	self checkLabel: label.
	instructions addLast: (Message
										selector: #jump:
										argument: label)! !

!AssemblerMethod methodsFor: 'assembling' stamp: 'eem 08/07/2014 11:24'!
jump: label if: cond
	self checkLabel: label.
	instructions addLast: (Message
										selector: #jump:if:
										arguments: {label. cond})! !

!AssemblerMethod methodsFor: 'assembling' stamp: 'eem 08/07/2014 11:24'!
jumpLong: label
	self checkLabel: label.
	instructions addLast: (Message
										selector: #jumpLong:
										argument: label)! !

!AssemblerMethod methodsFor: 'assembling' stamp: 'eem 08/07/2014 11:24'!
jumpLong: label if: cond
	self checkLabel: label.
	instructions addLast: (Message
										selector: #jumpLong:if:
										arguments: {label. cond})! !

!AssemblerMethod methodsFor: 'assembling' stamp: 'eem 08/07/2014 11:24'!
label: aString 
	instructions addLast: aString -> nil! !

!AssemblerMethod methodsFor: 'accessing' stamp: 'eem 3/20/2013 12:56'!
labels
	^instructions select: [:ea| ea isVariableBinding]! !

!AssemblerMethod methodsFor: 'assembling' stamp: 'eem 8/7/2014 11:23'!
literal: aLiteral
	literals addLast: aLiteral! !

!AssemblerMethod methodsFor: 'accessing' stamp: 'eem 2/22/2013 04:31'!
literals
	"Answer the value of literals"

	^ literals! !

!AssemblerMethod methodsFor: 'accessing' stamp: 'eem 2/22/2013 04:31'!
literals: anObject
	"Set the value of literals"

	literals := anObject! !

!AssemblerMethod methodsFor: 'initialize-release' stamp: 'eem 2/24/2013 13:17'!
literals: literalSequence instructions: instructionSequence numArgs: nArgs numTemps: nTemps frameSize: frameSizeNum primitive: primitiveIndex flag: flagBoolean signFlag: signFlagBoolean trailer: methodTrailer methodClass: class selector: aSelector compiledMethodClass: classOfCompiledMethod
	literals := literalSequence.
	instructions := instructionSequence.
	numArgs := nArgs.
	numTemps := nTemps.
	frameSize := frameSizeNum.
	primitive := primitiveIndex.
	flag := flagBoolean.
	signFlag := signFlagBoolean.
	trailer := methodTrailer.
	methodClass := class.
	selector := aSelector.
	compiledMethodClass := classOfCompiledMethod! !

!AssemblerMethod methodsFor: 'accessing' stamp: 'eem 2/22/2013 10:06'!
methodClass
	"Answer the value of methodClass"

	^ methodClass! !

!AssemblerMethod methodsFor: 'accessing' stamp: 'eem 2/22/2013 10:06'!
methodClass: anObject
	"Set the value of methodClass"

	methodClass := anObject! !

!AssemblerMethod methodsFor: 'accessing' stamp: 'eem 2/22/2013 04:31'!
numArgs
	"Answer the value of numArgs"

	^ numArgs! !

!AssemblerMethod methodsFor: 'accessing' stamp: 'eem 2/22/2013 04:31'!
numArgs: anObject
	"Set the value of numArgs"

	numArgs := anObject! !

!AssemblerMethod methodsFor: 'accessing' stamp: 'eem 2/22/2013 04:31'!
numTemps
	"Answer the value of numTemps"

	^ numTemps! !

!AssemblerMethod methodsFor: 'accessing' stamp: 'eem 2/22/2013 04:31'!
numTemps: anObject
	"Set the value of numTemps"

	numTemps := anObject! !

!AssemblerMethod methodsFor: 'assembling' stamp: 'eem 08/07/2014 11:24'!
popIntoLiteralVariableAtIndex: litIndex
	instructions addLast: (Message
										selector: #popIntoLiteralVariableAtIndex:
										argument: litIndex)! !

!AssemblerMethod methodsFor: 'accessing' stamp: 'eem 2/22/2013 04:31'!
primitive
	"Answer the value of primitive"

	^ primitive! !

!AssemblerMethod methodsFor: 'accessing' stamp: 'eem 3/14/2013 14:38'!
primitive: anObject
	"Set the value of primitive"

	primitive := anObject! !

!AssemblerMethod methodsFor: 'printing' stamp: 'eem 8/1/2014 20:46'!
printAccessorsForAssemblyOn: aStream
	| dummy |
	dummy := self class new.
	"Generate all the accessors; for now don't handle the trailer"
	#(			compiledMethodClass: methodClass: selector: primitive:
				numArgs: numTemps: frameSize: flag: signFlag: trailerData:)
		with: #(	compiledMethodClass methodClass selector primitive
				numArgs numTemps frameSize flag signFlag trailerData)
		do: [:setter :getter|
			(self perform: getter) ifNotNil:
				[:val|
				val ~= (dummy perform: getter) ifTrue:
					[aStream crtab; nextPutAll: setter; space; print: val; nextPut: $;]]]! !

!AssemblerMethod methodsFor: 'printing' stamp: 'eem 8/1/2014 20:51'!
printAsAssemblerOn: aStream
	aStream nextPut: $(; print: self class; nextPutAll: ' new'.

	self printAccessorsForAssemblyOn: aStream.
	self printLiteralsForAssemblyOn: aStream.
	self printInstructionsForAssemblyOn: aStream.

	aStream crtab; nextPutAll: #yourself; nextPut: $)! !

!AssemblerMethod methodsFor: 'printing' stamp: 'eem 3/15/2013 11:30'!
printInstructionsForAssemblyOn: aStream
	instructions do:
		[:messageOrLabel|
		aStream crtab.
		(self isLabel: messageOrLabel)
			ifTrue:
				[aStream nextPutAll: 'label: '; print: messageOrLabel key]
			ifFalse:
				[messageOrLabel numArgs = 0
					ifTrue: [aStream nextPutAll: messageOrLabel selector]
					ifFalse:
						[messageOrLabel selector keywords with: messageOrLabel arguments do:
							[:kwd :arg|
							aStream nextPutAll: kwd; space.
							aStream print: ((kwd = 'jump:' or: [kwd = 'blockSize:'])
												ifTrue: [self assert: (self isLabel: arg).
														arg key]
												ifFalse: [arg]); space].
						aStream skip: -1]].
		aStream nextPut: $;]! !

!AssemblerMethod methodsFor: 'printing' stamp: 'eem 3/15/2013 10:49'!
printLiteralsForAssemblyOn: aStream
	| literalsToDefine |
	literalsToDefine := literals copy asOrderedCollection.
	(literalsToDefine last isVariableBinding
	 and: [literalsToDefine last value = methodClass]) ifTrue:
		[literalsToDefine removeLast.
		 self assert: literalsToDefine last == selector.
		 literalsToDefine removeLast].
	literalsToDefine do:
		[:litOrBinding|
		aStream crtab.
		litOrBinding isLiteral
			ifTrue: [aStream nextPutAll: 'literal: '; print: litOrBinding]
			ifFalse: [aStream nextPutAll: 'binding: '; store: litOrBinding key].
		aStream nextPut: $;]! !

!AssemblerMethod methodsFor: 'assembling' stamp: 'eem 08/07/2014 11:24'!
pushClosureCopyNumCopiedValues: numCopied numArgs: numArgs blockSize: blockSize
	self checkLabel: blockSize.
	instructions addLast: (Message
										selector: #pushClosureCopyNumCopiedValues:numArgs:blockSize:
										arguments: {numCopied. numArgs. blockSize})! !

!AssemblerMethod methodsFor: 'assembling' stamp: 'eem 08/07/2014 11:24'!
pushConstantAtIndex: litIndex
	instructions addLast: (Message
										selector: #pushConstantAtIndex:
										argument: litIndex)! !

!AssemblerMethod methodsFor: 'assembling' stamp: 'eem 08/07/2014 11:24'!
pushLiteralVariableAtIndex: litIndex
	instructions addLast: (Message
										selector: #pushLiteralVariableAtIndex:
										argument: litIndex)! !

!AssemblerMethod methodsFor: 'assembling' stamp: 'eem 08/07/2014 11:24'!
pushSpecialConstant: aConstant
	instructions addLast: (Message
										selector: #pushSpecialConstant:
										argument: aConstant)! !

!AssemblerMethod methodsFor: 'accessing' stamp: 'eem 2/24/2013 13:16'!
selector
	"Answer the value of selector"

	^ selector! !

!AssemblerMethod methodsFor: 'accessing' stamp: 'eem 2/24/2013 13:16'!
selector: anObject
	"Set the value of selector"

	selector := anObject! !

!AssemblerMethod methodsFor: 'assembling-shorthand' stamp: 'eem 8/7/2014 11:19'!
send: aSelector
	self send: aSelector super: false numArgs: aSelector numArgs! !

!AssemblerMethod methodsFor: 'accessing' stamp: 'eem 2/22/2013 04:31'!
signFlag
	"Answer the value of signFlag"

	^ signFlag! !

!AssemblerMethod methodsFor: 'accessing' stamp: 'eem 2/22/2013 04:31'!
signFlag: anObject
	"Set the value of signFlag"

	signFlag := anObject! !

!AssemblerMethod methodsFor: 'assembling' stamp: 'eem 08/07/2014 11:24'!
storeIntoLiteralVariableAtIndex: litIndex
	instructions addLast: (Message
										selector: #storeIntoLiteralVariableAtIndex:
										argument: litIndex)! !

!AssemblerMethod methodsFor: 'assembling-shorthand' stamp: 'eem 8/7/2014 11:25'!
superSend: aSelector
	self send: aSelector super: true numArgs: aSelector numArgs! !

!AssemblerMethod methodsFor: 'accessing' stamp: 'eem 8/7/2014 11:08'!
trailer
	^trailer ifNil: [CompiledMethodTrailer empty]! !

!AssemblerMethod methodsFor: 'accessing' stamp: 'eem 2/22/2013 10:06'!
trailer: anObject
	"Set the value of trailer"

	trailer := anObject! !

!AssemblerMethod methodsFor: 'accessing' stamp: 'eem 3/15/2013 10:07'!
trailerData
	^trailer ifNotNil: [trailer rawData]! !

!AssemblerMethod methodsFor: 'accessing' stamp: 'eem 3/15/2013 10:07'!
trailerData: aByteArray
	self trailer: (CompiledMethodTrailer new method: aByteArray)! !

Object subclass: #BytecodeAssembler
	instanceVariableNames: 'code sizes encoder method index shouldNegateInstVarOffsetsMask'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MethodMassage-Kernel'!

!BytecodeAssembler commentStamp: 'eem 2/22/2013 11:12' prior: 0!
A BytecodeAssembler is something that converts an AssemblerMethod into a CompiledMethod.

Instance Variables
	code:		<AssemblerMethod>
	encoder:	<BytecodeEncoder>
	sizes:		<SequenceableCollection>

code
	- the input assembly method

encoder
	- the encoder that generates instructions

sizes
	- the size of each corresponding bytecode in code's instructions!

!BytecodeAssembler methodsFor: 'assembly-public' stamp: 'eem 2/24/2013 13:38'!
assemble: ass "<AssemblerMethod>"
	code := ass.
	sizes := Array new: code instructions size.
	self createEncoder.
	self nilLabelTargets.
	self sizePasses.
	self createMethod.
	self emitPass.
	^method! !

!BytecodeAssembler methodsFor: 'accessing' stamp: 'eem 2/23/2013 16:53'!
code
	^code! !

!BytecodeAssembler methodsFor: 'instance initialization' stamp: 'eem 8/7/2014 11:40'!
createEncoder
	"The signFlag defines the bytecode set of a method.
	 So create a dummy method with its sign flag to that of
	 code's, and instantiate the dummy method's encoderClass."
	| dummyMethod |
	encoder ifNil:
		[dummyMethod := CompiledMethod
							newMethod: 0
							header: (code signFlag ifTrue: [SmallInteger minVal] ifFalse: [0]).
		 encoder := dummyMethod encoderClass new].
	encoder
		initClass: code methodClass
		literalStream: (code literals isEmpty ifFalse:
						[self
							ensureLiteralsIncludeMethodClass: code literals
							isQuick: (self isQuickPrimitiveIndex: code primitive)]).
	^encoder! !

!BytecodeAssembler methodsFor: 'assembly' stamp: 'eem 2/23/2013 23:09'!
createMethod
	| header size |
	header := encoder
					computeMethodHeaderForNumArgs: code numArgs
					numTemps: code numTemps
					numLits: encoder allLiterals size
					primitive: code primitive.
	
	code flag ifTrue: "kind of crappy that there's no setter..."
		[header := header bitOr: (1 << 29)].
	size := sizes size = 0 ifTrue: [0] ifFalse: [self totalSizeFrom: 1 to: sizes size].
	method := code trailer
					createMethod: size
					class: code compiledMethodClass
					header: header.
	method needsFrameSize: code frameSize - method numTemps! !

!BytecodeAssembler methodsFor: 'assembly' stamp: 'eem 2/24/2013 13:47'!
emitPass
	| literals nLits emitter |

	literals := encoder allLiterals.
	nLits := literals size.
	emitter := BytecodeEmitter assembler: self encoder: encoder.

	1 to: nLits do: [:lit | method literalAt: lit put: (literals at: lit)].

	encoder streamToMethod: method.

	code instructions withIndexDo:
		[:messageOrLabel :i|
		(self isLabel: messageOrLabel) ifFalse:
			[index := i.
			 messageOrLabel sentTo: emitter]].

	"Need to create distinct literals for all literals in the method
	 *except* variable bindings and the methodClass literal.."
	1 to: method numLiterals - 2 do:
		[:litIndex| | lit |
		lit := method literalAt: litIndex.
		lit isVariableBinding ifFalse:
			[method
				literalAt: litIndex
				put: (lit isArray ifTrue: [lit deepCopy] ifFalse: [lit copy])]].

	"Need to create a distinct additional method state if it has one.
	 Avoid setting selector if already set (method penultimateLiteral notNil and not a metod properties)"
	method penultimateLiteral
		ifNil: [method selector: code selector]
		ifNotNil:
			[:plit|
			plit isMethodProperties ifTrue:
				[method
					penultimateLiteral: plit copy;
					selector: code selector]]
	! !

!BytecodeAssembler methodsFor: 'assembly-public' stamp: 'eem 8/1/2014 06:05'!
encoder: aBytecodeEncoder
	encoder := aBytecodeEncoder! !

!BytecodeAssembler methodsFor: 'private' stamp: 'eem 3/21/2013 06:46'!
ensureLiteralsIncludeMethodClass: literals isQuick: isQuick
	"Check whether literals includes the methodClass association (and hence by implication the penultimate literal).
	 If not, add a suitable one.  In the assembly case the literal will be missing."
	literals size > 0 ifTrue:
		[| last |
		isQuick ifTrue: "support Newspeak accessors"
			[self assert: literals size = 2.
			 ^literals].
		last := literals last.
		(last isVariableBinding
		and: [last value = code methodClass
		and: [last key == nil or: [last == code methodClass binding]]]) ifTrue:
			[^literals]].
	^literals, { code selector. code methodClass ifNotNil: [:class| class binding] }! !

!BytecodeAssembler methodsFor: 'accessing' stamp: 'eem 2/22/2013 15:01'!
index
	"For the sizer."
	^index! !

!BytecodeAssembler methodsFor: 'testing' stamp: 'eem 2/22/2013 12:47'!
isLabel: aMessageOrLookupKey
	^aMessageOrLookupKey isVariableBinding! !

!BytecodeAssembler methodsFor: 'testing' stamp: 'eem 2/23/2013 15:39'!
isQuickPrimitiveIndex: anInteger
	^anInteger between: 256 and: 519! !

!BytecodeAssembler methodsFor: 'assembly' stamp: 'eem 2/22/2013 14:29'!
nilLabelTargets
	code instructions do:
		[:messageOrLabel|
		(self isLabel: messageOrLabel) ifTrue:
			[messageOrLabel value: nil]]! !

!BytecodeAssembler methodsFor: 'testing' stamp: 'eem 2/24/2013 12:02'!
shouldUseLongAccessForInstVarOffset: offset
	"compute the mask of which inst var offsets need to be negated.
	 This is true of inst vars of contexts only."
	shouldNegateInstVarOffsetsMask ifNotNil:
		[:snivo| ^(snivo bitAnd: (1 bitShift: offset)) ~= 0].
	"The order in which this is done means that if the first offset is +ve all others will be too."
	shouldNegateInstVarOffsetsMask := 0.
	code methodClass instVarNamesAndOffsetsDo:
		[:n :o|
		o >= 0 ifTrue:
			[^(shouldNegateInstVarOffsetsMask bitAnd: (1 bitShift: offset)) ~= 0].
		o < 0 ifTrue:
			[shouldNegateInstVarOffsetsMask := shouldNegateInstVarOffsetsMask + (1 bitShift: -1 - o)]].
	^true! !

!BytecodeAssembler methodsFor: 'assembly' stamp: 'eem 2/24/2013 13:53'!
sizePasses
	| sizer instructions lastForwardJump i limit messageOrLabel |
	sizer := BytecodeSizer assembler: self encoder: encoder.
	"Forward pass that sizes all except forward jumps."
	i := 0.
	limit := (instructions := code instructions) size.
	[(i := i + 1) <= limit] whileTrue:
		[messageOrLabel := instructions at: i.
		sizes
			at: i
			put: ((self isLabel: messageOrLabel)
					ifTrue: [0]
					ifFalse: [index := i.
							(messageOrLabel sentTo: sizer)
								ifNil: [lastForwardJump := i. nil]
								ifNotNil: [:size| size]])].
	lastForwardJump ifNil: [^self].
	"Backward pass that sizes forward jumps."
	i := lastForwardJump + 1.
	[(i := i - 1) >= 1] whileTrue:
		[(sizes at: i) ifNil:
			[index := i.
			 sizes
				at: i
				put: ((instructions at: i) sentTo: sizer)]]! !

!BytecodeAssembler methodsFor: 'assembly' stamp: 'eem 2/22/2013 14:33'!
spanToLabel: label "<Association>"
	| labelIndex |
	labelIndex := label value.
	labelIndex ifNil:
		[labelIndex := code instructions identityIndexOf: label.
		 label value: labelIndex].
	^(sizes at: labelIndex) ifNotNil:
		[self totalSizeFrom: index + 1 to: labelIndex]! !

!BytecodeAssembler methodsFor: 'assembly' stamp: 'eem 2/22/2013 17:05'!
totalSizeFrom: startPC to: targetPC
	| sum |
	sum := 0.
	^(sizes at: targetPC) ifNotNil:
		[startPC <= targetPC
			ifTrue:
				[startPC to: targetPC do:
					[:i| sum := sum + (sizes at: i)]]
			ifFalse:
				[targetPC to: startPC - 1 do:
					[:i|
					"If a nil is found it will be for a forward branch in the first pass, and
					 the total size will be being requested by a backward branch.  Simply
					 answer anything negative.  All negative branches are long."
					(sizes at: i)
						ifNil: [^-1]
						ifNotNil: [:size| sum := sum - size]]].
		 sum]! !

Object subclass: #BytecodeCoverer
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MethodMassage-Coverage'!

!BytecodeCoverer class methodsFor: 'utilities' stamp: 'eem 3/20/2013 17:33'!
unusedBytecodeSendsEnabled
	^(Smalltalk specialObjectsArray at: 58 ifAbsent: []) == #unusedBytecode! !

!BytecodeCoverer methodsFor: 'private' stamp: 'eem 3/20/2013 17:33'!
checkUnusedBytecodeSendsEnabled
	self class unusedBytecodeSendsEnabled ifFalse:
		[self error: 'unusedBytecode sends are not enabled.  continuting will likely crash the VM']! !

!BytecodeCoverer methodsFor: 'covering' stamp: 'eem 3/20/2013 17:30'!
coverMethod: aMethod
	| labels basicBlockStarts unusedBytecode |
	self checkUnusedBytecodeSendsEnabled.
	labels := (BasicBlockLabellingDisassembler new disassemble: aMethod) labels.
	basicBlockStarts := Dictionary new.
	unusedBytecode := aMethod encoderClass unusedBytecode.
	self assert: unusedBytecode notNil.
	labels do:
		[:assoc| | pc |
		pc := assoc value.
		basicBlockStarts at: pc put: (aMethod at: pc).
		aMethod at: pc put: unusedBytecode].
	aMethod
		propertyValueAt: #coverage put: basicBlockStarts;
		voidCogVMState! !

!BytecodeCoverer methodsFor: 'covering' stamp: 'eem 3/20/2013 17:47'!
uncoverMethod: aMethod
	| coverage |
	coverage := aMethod propertyValueAt: #coverage.
	coverage ifNil: [^self].
	coverage keysAndValuesDo:
		[:pc :bytecode|
		aMethod at: pc put: bytecode].
	aMethod removeProperty: #coverage.
	(aMethod penultimateLiteral isMethodProperties
	 and: [aMethod penultimateLiteral isEmpty]) ifTrue:
		[aMethod penultimateLiteral: aMethod selector]! !

Object subclass: #BytecodeDisassembler
	instanceVariableNames: 'instructions labels labelling instrs thePC method'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MethodMassage-Kernel'!

BytecodeDisassembler subclass: #BasicBlockLabellingDisassembler
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MethodMassage-Coverage'!

!BasicBlockLabellingDisassembler methodsFor: 'disassembly' stamp: 'eem 3/20/2013 14:02'!
label
	"Override to add a label for the first bytecode pc.
	 Be careful to skip any initial callPrimitive bytecode."
	method isQuick ifFalse:
		[| startPC primCode|
		 startPC := (method primitive ~= 0
					 and: [(primCode := method encoderClass callPrimitiveCode) notNil])
						ifTrue: [method initialPC + ((InstructionStream on: method) bytecodeSize: primCode)]
						ifFalse: [method initialPC].
		 labels at: startPC put: true].
	super label! !

!BasicBlockLabellingDisassembler methodsFor: 'instruction decoding' stamp: 'eem 3/20/2013 13:36'!
pushClosureCopyNumCopiedValues: numCopied numArgs: numArgs blockSize: blockSize
	"Override to add a label for the start of a block."
	labelling ifTrue:
		[labels at: instrs pc put: true].
	super pushClosureCopyNumCopiedValues: numCopied numArgs: numArgs blockSize: blockSize! !

!BytecodeDisassembler methodsFor: 'disassembly' stamp: 'eem 2/24/2013 13:18'!
disassemble: aMethod
	| endPC |
	method := aMethod.
	labels := Array new: aMethod size + 1 withAll: false.
	instructions := Array new: aMethod size + 1.
	instrs := InstructionStream on: aMethod.
	endPC := aMethod endPC.
	labelling := true.
	[(thePC := instrs pc) <= endPC] whileTrue:
		[instrs interpretNextInstructionFor: self].
	self label.
	labelling := false.
	instrs pc: aMethod initialPC.
	[(thePC := instrs pc) <= endPC] whileTrue:
		[instrs interpretNextInstructionFor: self].
	^AssemblerMethod
		literals: aMethod literals
		instructions: self instructionSequence
		numArgs: aMethod numArgs
		numTemps: aMethod numTemps
		frameSize: aMethod frameSize
		primitive: aMethod primitive
		flag: aMethod flag
		signFlag: aMethod signFlag
		trailer: aMethod trailer
		methodClass: aMethod methodClass
		selector: aMethod selector
		compiledMethodClass: aMethod class
	
	"BytecodeDisassembler new disassemble: BytecodeDisassembler >> #jump:if:"! !

!BytecodeDisassembler methodsFor: 'disassembly' stamp: 'eem 2/22/2013 12:48'!
doesNotUnderstand: aMessage
	(InstructionClient includesSelector: aMessage selector) ifFalse:
		[^super doesNotUnderstand: aMessage].
	aMessage lookupClass: nil.
	instructions at: thePC put: aMessage! !

!BytecodeDisassembler methodsFor: 'disassembly' stamp: 'eem 2/22/2013 04:35'!
instructionSequence
	^Array streamContents:
		[:s|
		labels with: instructions do:
			[:label :inst|
			inst ifNotNil:
				[label ~~ false ifTrue:
					[s nextPut: label].
				s nextPut: inst]]]! !

!BytecodeDisassembler methodsFor: 'instruction decoding' stamp: 'eem 2/21/2013 16:46'!
jump: offset
	"Print the Unconditional Jump bytecode."

	labelling
		ifTrue:
			[labels at: instrs pc + offset put: true]
		ifFalse:
			[instructions
				at: thePC
				put: (Message
						selector: #jump:
						argument: (labels at: instrs pc + offset))]! !

!BytecodeDisassembler methodsFor: 'instruction decoding' stamp: 'eem 2/21/2013 16:46'!
jump: offset if: condition
	"Disassemble the Conditional Jump bytecode."

	labelling
		ifTrue:
			[labels at: instrs pc + offset put: true]
		ifFalse:
			[instructions
				at: thePC
				put: (Message
						selector: #jump:if:
						arguments: { labels at: instrs pc + offset. condition })]! !

!BytecodeDisassembler methodsFor: 'disassembly' stamp: 'eem 3/20/2013 12:41'!
label
	"N.B. A Label is an Association so that later its pc can be sorted in its value field."
	| n |
	n := 0.
	labels doWithIndex:
		[:l :i|
		l ifTrue:
			[labels at: i put: (Association new key: 'L', (n := n + 1) printString value: i)]]! !

!BytecodeDisassembler methodsFor: 'instruction decoding' stamp: 'eem 3/14/2013 16:58'!
popIntoLiteralVariable: aLiteral
	| litIndex |
	litIndex := method encoderClass literalIndexOfBytecodeAt: thePC in: method.
	litIndex ifNil: [self error: 'could not decode literal index'].
	instructions
		at: thePC
		put: (Message
				selector: #popIntoLiteralVariableAtIndex:
				argument: litIndex)! !

!BytecodeDisassembler methodsFor: 'instruction decoding' stamp: 'eem 2/22/2013 04:42'!
pushClosureCopyNumCopiedValues: numCopied numArgs: numArgs blockSize: blockSize
	labelling
		ifTrue:
			[labels at: instrs pc + blockSize put: true]
		ifFalse:
			[instructions
				at: thePC
				put: (Message
						selector: #pushClosureCopyNumCopiedValues:numArgs:blockSize:
						arguments: { numCopied. numArgs. labels at: instrs pc + blockSize })]! !

!BytecodeDisassembler methodsFor: 'instruction decoding' stamp: 'eem 3/14/2013 16:40'!
pushConstant: aLiteral
	instructions
		at: thePC
		put: ((method encoderClass literalIndexOfBytecodeAt: thePC in: method)
				ifNil: [Message
						selector: #pushSpecialConstant:
						argument: aLiteral]
				ifNotNil:
					[:index|
					Message
						selector: #pushConstantAtIndex:
						argument: index])! !

!BytecodeDisassembler methodsFor: 'instruction decoding' stamp: 'eem 3/14/2013 16:59'!
pushLiteralVariable: aLiteral
	| litIndex |
	litIndex := method encoderClass literalIndexOfBytecodeAt: thePC in: method.
	litIndex ifNil: [self error: 'could not decode literal index'].
	instructions
		at: thePC
		put: (Message
				selector: #pushLiteralVariableAtIndex:
				argument: litIndex)! !

!BytecodeDisassembler methodsFor: 'instruction decoding' stamp: 'eem 3/14/2013 16:59'!
storeIntoLiteralVariable: aLiteral
	| litIndex |
	litIndex := method encoderClass literalIndexOfBytecodeAt: thePC in: method.
	litIndex ifNil: [self error: 'could not decode literal index'].
	instructions
		at: thePC
		put: (Message
				selector: #storeIntoLiteralVariableAtIndex:
				argument: litIndex)! !

!MethodContext methodsFor: '*MethodMassage-coverage' stamp: 'bwesterg 9/30/2013 17:32'!
unusedBytecode
	"Handle unusedBytecode by replacing the bytecode with the
	 correct one found in the coverage property and continuing.
	 Continue via wait/signal since return would push a result."
	| coverage semaphore process originalBytecode |
	self assert: (method at: pc) = method encoderClass unusedBytecode.
	coverage := method propertyValueAt: #coverage.
	self assert: coverage notNil.
	originalBytecode := coverage at: pc ifAbsent: [nil].
	self assert: originalBytecode notNil.
	semaphore := Semaphore new.
	process := Processor activeProcess.

	[method at: pc put: originalBytecode.
	 coverage removeKey: pc ifAbsent: [].
	 process suspendedContext unwindTo: self.
	 process suspendedContext: self.
	 semaphore signal] fork.

	semaphore wait! !

!CompiledMethodTrailer methodsFor: '*MethodMassage-raw data' stamp: 'eem 3/15/2013 10:02'!
rawData
	^(ByteArray new: size)
		replaceFrom: 1
		to: size
		with: method
		startingAt: method size - size + 1! !

TestCase subclass: #BytecodeAssemblerTests
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MethodMassage-Tests'!

!BytecodeAssemblerTests class methodsFor: 'testing' stamp: 'eem 4/9/2013 12:14'!
expectedFailures
	(thisContext method respondsTo: #usesAlternateBytecodeSet) ifTrue:
		[^#()].
	^#(testDuplicateNewsqueakV4Methods)! !

!BytecodeAssemblerTests class methodsFor: 'utilities' stamp: 'eem 3/15/2013 13:04'!
fixupMethodClassBindings
	"Work-around the Monticello bug that results in not sharing the binding in Smalltalk as a method's methodClassAssociation."
	"self fixupMethodClassBindings"
	self systemNavigation allSelect:
		[:m|
		(m methodClass isMeta not
		 and: [m methodClassAssociation key isNil]) ifTrue:
			[m methodClassAssociation: m methodClass binding].
		false]! !

!BytecodeAssemblerTests methodsFor: 'test support' stamp: 'eem 2/23/2013 16:08'!
contextPartMethodWithMostOf: binaryBlock
	^(Array streamContents: [:s| ContextPart withAllSubclassesDo: [:sc| s nextPutAll: sc methodDict values]])
		inject: ContextPart methodDict anyOne
		into:
			[:m1 :m2|
			((1 to: m1 methodClass instSize) select: [:i| binaryBlock value: m1 value: i]) size >=
			((1 to: m2 methodClass instSize) select: [:i| binaryBlock value: m2 value: i]) size
				ifTrue: [m1]
				ifFalse: [m2]]! !

!BytecodeAssemblerTests methodsFor: 'accessing' stamp: 'eem 2/23/2013 23:12'!
defaultTimeout
	"Answer the default timeout to use for tests in this test case.
	The timeout is a value in seconds."

	^30 "seconds"! !

!BytecodeAssemblerTests methodsFor: 'testing' stamp: 'eem 4/9/2013 12:17'!
expectedFailures
	(thisContext method respondsTo: #usesAlternateBytecodeSet) ifTrue:
		[^#()].
	^#(testDuplicateNewsqueakV4Methods)! !

!BytecodeAssemblerTests methodsFor: 'test support' stamp: 'eem 3/20/2013 12:48'!
testAsAssemblerFor: aMethod
	"self new testAsAssemblerFor: self >> #testAsAssemblerFor:"
	| assemblerMethod assembler reassembledMethod |
	assemblerMethod := BytecodeDisassembler new disassemble: aMethod.
	assembler := String streamContents: [:s| assemblerMethod printAsAssemblerOn: s].
	self shouldnt: [reassembledMethod := Compiler evaluate: assembler]
		raise: Error.
	self assert: assemblerMethod compiledMethodClass equals: reassembledMethod compiledMethodClass.
	self assert: assemblerMethod flag equals: reassembledMethod flag.
	self assert: assemblerMethod frameSize equals: reassembledMethod frameSize.
	self assert: assemblerMethod methodClass equals: reassembledMethod methodClass.
	self assert: assemblerMethod numArgs equals: reassembledMethod numArgs.
	self assert: assemblerMethod numTemps equals: reassembledMethod numTemps.
	self assert: assemblerMethod primitive equals: reassembledMethod primitive.
	self assert: assemblerMethod selector equals: reassembledMethod selector.
	self assert: assemblerMethod signFlag equals: reassembledMethod signFlag.
	1 to: reassembledMethod literals size do:
		[:i|
		self assert: (assemblerMethod literals at: i)
			equals: (reassembledMethod literals at: i)].
	assemblerMethod instructions
		ifNil: [self assert: reassembledMethod instructions isNil]
		ifNotNil:
			[:insts|
			 self assert: insts size equals: reassembledMethod instructions size.
			 "Nil the labels' values for analogousCodeTo: comparison below; the values
			  are usefully the bytecode pcs of the labels, but this fails the comparisons."
			 insts do: [:inst| inst isVariableBinding ifTrue: [inst value: nil]].
			 insts
				with: reassembledMethod instructions
				do: [:old :new|
					old isVariableBinding
						ifTrue: [self assert: old key = new key]
						ifFalse: [self assert: (old analogousCodeTo: new)]]].
	self assert: aMethod equals: (BytecodeAssembler new assemble: reassembledMethod)! !

!BytecodeAssemblerTests methodsFor: 'tests' stamp: 'eem 3/15/2013 08:57'!
testAsAssemblerForThisPackage
	(PackageOrganizer default packageOfClass: self class) actualMethodsDo:
		[:m| self testAsAssemblerFor: m]! !

!BytecodeAssemblerTests methodsFor: 'test support' stamp: 'eem 3/20/2013 13:47'!
testBasicBlockLabelsFor: aMethod
	"self new testBasicBlockLabelsFor: self >> #testBasicBlockLabelsFor:"
	| labels stream endPC |
	labels := (BasicBlockLabellingDisassembler new disassemble: aMethod) labels.
	aMethod isQuick ifTrue:
		[^self assert: labels isEmpty].
	self assert: (labels select: [:l| aMethod initialPC = l value]) size = 1.
	aMethod embeddedBlockClosures do:
		[:closure|
		self assert: (labels select: [:l| closure startpc = l value]) size = 1].
	stream := InstructionStream on: aMethod.
	endPC := aMethod endPC.
	[stream pc <= endPC] whileTrue:
		[stream interpretJump
			ifNotNil: [:dist| self assert: (labels select: [:l| stream pc + dist = l value]) size = 1]
			ifNil: [stream pc: (stream nextPc: (aMethod at: stream pc))]]! !

!BytecodeAssemblerTests methodsFor: 'tests' stamp: 'eem 3/20/2013 13:42'!
testBasicBlockLabelsForThisPackage
	(PackageOrganizer default packageOfClass: self class) actualMethodsDo:
		[:m| self testBasicBlockLabelsFor: m]! !

!BytecodeAssemblerTests methodsFor: 'tests' stamp: 'eem 2/22/2013 11:35'!
testDisassembleBytecodeAssemblerMethods
	self testDisassembleMethodsFor: BytecodeAssembler! !

!BytecodeAssemblerTests methodsFor: 'test support' stamp: 'eem 2/22/2013 11:38'!
testDisassembleMethodsFor: aBehavior
	aBehavior selectors asArray sort do:
		[:sel| | method |
		method := aBehavior compiledMethodAt: sel.
		self shouldnt: [BytecodeDisassembler new disassemble: method]
			raise: Error]! !

!BytecodeAssemblerTests methodsFor: 'tests' stamp: 'eem 2/23/2013 15:49'!
testDuplicateAllInstructionStreamMethods
	InstructionStream withAllSuperclasses, InstructionStream allSubclasses do:
		[:isc|
		self testDuplicateMethodsFor: isc
			if: [:method| true]]! !

!BytecodeAssemblerTests methodsFor: 'tests' stamp: 'eem 2/23/2013 15:40'!
testDuplicateBytecodeAssemblerMethods
	self testDuplicateMethodsFor: BytecodeAssembler
		if: [:method| true]! !

!BytecodeAssemblerTests methodsFor: 'test support' stamp: 'eem 2/23/2013 17:54'!
testDuplicateMethodsFor: aBehavior if: methodSelectBlock
	aBehavior selectors asArray sort do:
		[:sel| | method disassembly assembly |
		method := aBehavior compiledMethodAt: sel.
		(methodSelectBlock value: method) ifTrue:
			[disassembly := BytecodeDisassembler new disassemble: method.
			 assembly := BytecodeAssembler new assemble: disassembly.
			"method indexOfDifferenceWith: assembly"
			 self assert: method equals: assembly]]! !

!BytecodeAssemblerTests methodsFor: 'tests' stamp: 'eem 2/23/2013 21:13'!
testDuplicateNewsqueakV4Methods
	Smalltalk allClasses do:
		[:class|
		self testDuplicateMethodsFor: class
			if: [:method| method usesAlternateBytecodeSet].
		self testDuplicateMethodsFor: class class
			if: [:method| method usesAlternateBytecodeSet]]! !

!BytecodeAssemblerTests methodsFor: 'tests' stamp: 'eem 4/9/2013 12:13'!
testDuplicateSqueakV3Methods
	"AndreasProfiler spyOn: [self new testDuplicateSqueakV3Methods]"
	| filter |
	filter := (thisContext method respondsTo: #usesAlternateBytecodeSet)
				ifTrue: [[:method| method usesAlternateBytecodeSet not]]
				ifFalse: [[:method| true]].
	Smalltalk allClasses do:
		[:class|
		self testDuplicateMethodsFor: class if: filter.
		self testDuplicateMethodsFor: class class if: filter]! !

!BytecodeAssemblerTests methodsFor: 'test support' stamp: 'eem 3/20/2013 13:15'!
testLabelsFor: aMethod
	"self new testLabelsFor: self >> #testLabelsFor:"
	| labels stream endPC |
	labels := (BytecodeDisassembler new disassemble: aMethod) labels.
	stream := InstructionStream on: aMethod.
	endPC := aMethod endPC.
	[stream pc <= endPC] whileTrue:
		[stream interpretJump
			ifNotNil: [:dist| self assert: (labels select: [:l| stream pc + dist = l value]) size = 1]
			ifNil: [stream pc: (stream nextPc: (aMethod at: stream pc))]]! !

!BytecodeAssemblerTests methodsFor: 'tests' stamp: 'eem 3/20/2013 13:17'!
testLabelsForThisPackage
	(PackageOrganizer default packageOfClass: self class) actualMethodsDo:
		[:m| self testLabelsFor: m]! !

!BytecodeAssemblerTests methodsFor: 'tests' stamp: 'eem 8/7/2014 11:51'!
testManualAssembly
	self
		assert:
			(AssemblerMethod new
				methodClass: Object; selector: #printOn:; numArgs: 1; numTemps: 2;
				pushReceiver;
				send: #class;
				send: #name;
				popIntoTemporaryVariable: 1;
				pushTemporaryVariable: 0;
				doDup;
				pushTemporaryVariable: 1;
				send: #first;
				send: #isVowel;
				jump: 'L1' if: false;
				pushConstant: 'an ';
				jump: 'L2';
			label: 'L1';
				pushConstant: 'a ';
			label: 'L2';
				send: #nextPutAll:;
				doPop;
				pushTemporaryVariable: 1;
				send: #nextPutAll:;
				doPop;
				methodReturnReceiver;
				assemble) decompile printString
		equals: (Object >>#printOn:) decompile printString! !

TestCase subclass: #BytecodeCoverageTests
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MethodMassage-Tests'!

!BytecodeCoverageTests methodsFor: 'tests support' stamp: 'eem 3/21/2013 14:45'!
coverTestBasicBlockLabelsForMethod

	(PackageOrganizer default packageOfClass: self class) actualMethodsDo:
		[:m| self testBasicBlockLabelsFor: m]! !

!BytecodeCoverageTests methodsFor: 'initialize-release' stamp: 'eem 3/20/2013 16:11'!
setUp
	self class recompile: #testBasicBlockLabelsFor: from: BytecodeAssemblerTests! !

!BytecodeCoverageTests methodsFor: 'initialize-release' stamp: 'eem 3/20/2013 16:13'!
tearDown
	self class removeSelectorSilently: #testBasicBlockLabelsFor:! !

!BytecodeCoverageTests methodsFor: 'tests' stamp: 'eem 3/21/2013 14:54'!
testCoverageOfTestBasicBlockLabelsForMethod
	"[| me |
	me := self new.
	[me setUp; testCoverageOfTestBasicBlockLabelsForMethod]
		ensure: [me tearDown]] timeToRun"
	"[| me |
	me := self new.
	[me setUp; coverTestBasicBlockLabelsForMethod]
		ensure: [me tearDown]] timeToRun"
	| theMethod |
	theMethod := self class >> #testBasicBlockLabelsFor:.
	BytecodeCoverer new coverMethod: theMethod.
	self assert: (theMethod propertyValueAt: #coverage) notNil.
	self assert: (theMethod propertyValueAt: #coverage) notEmpty.

	self coverTestBasicBlockLabelsForMethod.

	self assert: ((theMethod) propertyValueAt: #coverage) isEmpty! !

!BytecodeEncoder class methodsFor: '*MethodMassage-assembly' stamp: 'eem 8/7/2014 11:13'!
assembler
	"Answer an assembler for a method that uses the receiver's instruction set."
	^AssemblerMethod new
		encoder: self new;
		yourself! !

!BytecodeEncoder methodsFor: '*MethodMassage-literal access' stamp: 'eem 2/23/2013 23:04'!
litIndexOrNilFor: aLiteral
	| index |
	index := aLiteral isVariableBinding
				ifTrue: [literalStream originalContents identityIndexOf: aLiteral ifAbsent: [^nil]]
				ifFalse: [literalStream originalContents indexOf: aLiteral ifAbsent: [^nil]].
	^index <= literalStream position ifTrue:
		[index - 1]! !

!EncoderForV3 class methodsFor: '*MethodMassage-disassembly' stamp: 'eem 5/19/2014 17:21'!
literalIndexOfBytecodeAt: pc in: aCompiledMethod
	"32-63 		001iiiii 						Push Literal Constant #iiiii
	 64-95 		010iiiii 						Push Literal Variable #iiiii
	 128 		10000000 jjkkkkkk 		Push (Receiver Variable, Temporary Location, Literal Constant, Literal Variable) [jj] #kkkkkk
	129 		10000001 jjkkkkkk 		Store (Receiver Variable, Temporary Location, Illegal, Literal Variable) [jj] #kkkkkk
	130 		10000010 jjkkkkkk 		Pop and Store (Receiver Variable, Temporary Location, Illegal, Literal Variable) [jj] #kkkkkk
	132 		10000100 iiijjjjj kkkkkkkk 	(Send, Send Super, Push Receiver Variable, Push Literal Constant, Push Literal Variable, Store Receiver Variable, Store-Pop Receiver Variable, Store Literal Variable)[iii] #kkkkkkkk jjjjj (for sends jjjjj = numArgs)"
	| byte type |
	byte := aCompiledMethod at: pc.
	byte <= 95 ifTrue:
		[^byte >= 32 ifTrue:
			[byte bitAnd: 16r1F]].
	(byte between: 128 and: 130) ifTrue:
		[^(aCompiledMethod at: pc + 1) bitAnd: 16r3F].
	byte = 132 ifTrue:
	[type := (aCompiledMethod at: pc + 1) bitShift: -5.
	 (type = 3 or: [type = 4 or: [type = 7]]) ifTrue:
		[^aCompiledMethod at: pc + 2]].
	^nil! !

!CompiledMethod methodsFor: '*MethodMassage-disassembly' stamp: 'eem 3/15/2013 13:16'!
disassemble
	^BytecodeDisassembler new disassemble: self! !

!CompiledMethod methodsFor: '*MethodMassage-debugging' stamp: 'eem 2/23/2013 16:30'!
indexOfDifferenceWith: method
	"Answer whether the receiver implements the same code as the argument, method.
	 Here ``same code'' means that if the receiver's source is compiled with the same
	 compiler it should produce the same sequence of bytecodes and literals, same
	 trailer and same properties.  Hence this definition of #= (only one of many plausible
	 definitions) can be used to quickly identify changes in the compiler's output."
	| numLits |
	method isCompiledMethod ifFalse: [^1].
	self size = method size ifFalse: [^-2].
	self header = method header ifFalse: [^-3]. "N.B. includes numLiterals comparison."
	self initialPC to: self endPC do:
		[:i | (self at: i) = (method at: i) ifFalse: [^i]].
	numLits := self numLiterals.
	1 to: numLits do:
		[:i| | lit1 lit2 |
		lit1 := self literalAt: i.
		lit2 := method literalAt: i.
		(lit1 == lit2 or: [lit1 literalEqual: lit2]) ifFalse:
			[(i = 1 and: [#(117 120) includes: self primitive])
				ifTrue:
					[lit1 isArray
						ifTrue:
							[(lit2 isArray and: [lit1 allButLast = lit2 allButLast]) ifFalse:
								[^i]]
						ifFalse: "ExternalLibraryFunction"
							[(lit1 analogousCodeTo: lit2) ifFalse:
								[^i]]]
				ifFalse:
					[i = (numLits - 1)
						ifTrue: "properties"
							[(self properties analogousCodeTo: method properties)
								ifFalse: [^i]]
						ifFalse: "last literal (methodClassAssociation) of class-side methods is not unique"
							[(i = numLits
							 and: [lit1 isVariableBinding
							 and: [lit2 isVariableBinding
							 and: [lit1 key == lit2 key
							 and: [lit1 value == lit2 value]]]]) ifFalse:
								[^i]]]]].
	^0! !

!Encoder methodsFor: '*MethodMassage-instance initialization' stamp: 'eem 8/7/2014 11:44'!
initClass: targetBehavior literalStream: literalsOrNil
	cue := CompilationCue class: targetBehavior.
	class := targetBehavior.
	optimizedSelectors := Set new.
	literalsOrNil
		ifNil:
			[literalStream := WriteStream on: (Array new: 32).
			 addedSelectorAndMethodClassLiterals := false]
		ifNotNil:
			[literalStream := (WriteStream on: literalsOrNil asArray from: 1 to: literalsOrNil size) setToEnd.
			 addedSelectorAndMethodClassLiterals := true]! !

!EncoderForNewsqueakV4 class methodsFor: '*MethodMassage-disassembly' stamp: 'eem 5/19/2014 17:18'!
literalIndexOfBytecodeAt: pc in: aCompiledMethod
	"	16-31		0001 i i i i				Push Literal Variable #iiii
	 	32-63		001 i i i i i				Push Literal #iiiii
	 *	224		11100000	aaaaaaaa	Extend A (Ext A = Ext A prev * 256 + Ext A)
	 *	227		11100011	i i i i i i i i	Push Literal Variable #iiiiiiii (+ Extend A * 256)
	 *	228		11100100	i i i i i i i i	Push Literal #iiiiiiii (+ Extend A * 256)
	 *	233		11101001	i i i i i i i i	Store Literal Variable #iiiiiiii (+ Extend A * 256)
	 *	236		11101100	i i i i i i i i	Pop and Store Literal Variable #iiiiiiii (+ Extend A * 256)"
	| byte |
	byte := aCompiledMethod at: pc.
	byte <= 63 ifTrue:
		[byte >= 32 ifTrue:
			[^byte bitAnd: 16r1F].
		 ^byte >= 16 ifTrue:
			[byte bitAnd: 16rF]].
	(#(227 228 233 236) includes: byte) ifTrue:
		[^aCompiledMethod at: pc + 1].
	(byte = 224 "Ext A"
	 and: [#(227 228 233 236) includes: (aCompiledMethod at: pc + 2)]) ifTrue:
		[^(aCompiledMethod at: pc + 1) * 256 + (aCompiledMethod at: pc + 3)].
	^nil! !

!EncoderForSistaV1 class methodsFor: '*MethodMassage-disassembly' stamp: 'eem 5/19/2014 17:18'!
literalIndexOfBytecodeAt: pc in: aCompiledMethod
	"	16-31		0001 i i i i				Push Literal Variable #iiii
	 	32-63		001 i i i i i				Push Literal #iiiii
	 *	224		11100000	aaaaaaaa	Extend A (Ext A = Ext A prev * 256 + Ext A)
	 *	227		11100011	i i i i i i i i	Push Literal Variable #iiiiiiii (+ Extend A * 256)
	 *	228		11100100	i i i i i i i i	Push Literal #iiiiiiii (+ Extend A * 256)
	 *	241		11110001	iiiiiiii		Pop and Store Literal Variable #iiiiiiii (+ Extend A * 256)
	 *	244		11110100	iiiiiiii		Store Literal Variable #iiiiiiii (+ Extend A * 256)"
	| byte |
	byte := aCompiledMethod at: pc.
	byte <= 63 ifTrue:
		[byte >= 32 ifTrue:
			[^byte bitAnd: 16r1F].
		 ^byte >= 16 ifTrue:
			[byte bitAnd: 16rF]].
	(#(227 228 241 244) includes: byte) ifTrue:
		[^aCompiledMethod at: pc + 1].
	(byte = 224 "Ext A"
	 and: [#(227 228 241 244) includes: (aCompiledMethod at: pc + 2)]) ifTrue:
		[^(aCompiledMethod at: pc + 1) * 256 + (aCompiledMethod at: pc + 3)].
	^nil! !
AssemblerEncoderInterface initialize!
